// empty.h
//
// Mark Johnson, 30th August 2001
//
// Contents:
//
//   class tree_slash_label
//   stree
//   empty_tree
//   adjunction_site
//
//   class eedge
//   empty_prec_rec

#ifndef EMPTY_H
#define EMPTY_H

#include <algorithm>
#include <functional>
#include <iostream>
#include <map>
#include <set>

#include "heads.h"
#include "parseval.h"
#include "tree.h"
#include "utility.h"

typedef std::set<size_t> sI;  // sI is a set of indices

class tree_slash_label : public tree_label {
public:
  sI slashes;

  tree_slash_label(const tree_label& tree_label_) : tree_label(tree_label_) { }

  tree_slash_label(const symbol cat_ = symbol::undefined(), size_t index_ = 0)
    : tree_label(cat_, index_) { };
};

std::ostream& operator<< (std::ostream& os, const tree_slash_label& label);

typedef tree_node<tree_slash_label> stree;

stree* tree_stree(const tree* t);


///////////////////////////////////////////////////////////////////////////
//                                                                       //
//                              spec_prec_rec                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////


typedef std::pair<int,int> II;
typedef std::pair<II,int> III;

template <typename Tree_Edges, typename tree_type, typename edge_label_type>
void spec_prec_rec(const Tree_Edges& tree_edges,
		   const tree_type* t1, const tree_type* t2, 
		   std::map<edge_label_type,parseval>& counts) // map from edge labels to count tuple
{
  typedef typename Tree_Edges::second_argument_type sE;
  typedef typename sE::value_type edge_type;   // type of edges

  sE e1s, e2s;
  tree_edges(t1, e1s);
  tree_edges(t2, e2s);

  typedef typename edge_type::first_type pos_type;
  typedef std::set<pos_type> sI;
  typedef std::pair<sI,sI> sIsI;
  typedef std::map<edge_label_type,sIsI> EL_sIsI;

  EL_sIsI el_s;
  cforeach (typename sE, e1it, e1s) 
    el_s[e1it->second].first.insert(e1it->first);
  cforeach (typename sE, e2it, e2s)
    el_s[e2it->second].second.insert(e2it->first);

  cforeach (typename EL_sIsI, it, el_s) {
    sI intersection;
    counts[it->first].ngold += it->second.first.size();
    counts[it->first].ntest += it->second.second.size();
    set_intersection(it->second.first, it->second.second, inserter(intersection));
    counts[it->first].ncommon += intersection.size();
  } 
}



///////////////////////////////////////////////////////////////////////////
//                                                                       //
//                    eedge -- edges for empty nodes                     //
//                                                                       //
///////////////////////////////////////////////////////////////////////////

struct eedge_label {
  symbol cat;       // NP, S, etc
  symbol type;      // 0, *, *T*, etc

  eedge_label(symbol cat_ = symbol::undefined(), symbol type_ = symbol::undefined())
    : cat(cat_), type(type_) { }

  bool operator== (const eedge_label& e) {
    return cat == e.cat && type == e.type;
  }

  bool operator< (const eedge_label& e) const {
    return cat < e.cat || (cat == e.cat && type < e.type);
  }
};

inline std::ostream& operator<< (std::ostream& os, const eedge_label& e) {
  return os << e.cat << ' ' << e.type;
}

namespace EXT_NAMESPACE {
  template <> struct hash<eedge_label> {
    size_t operator() (const eedge_label& e) const {
      return size_t(hash<symbol>()(e.cat) >> 1) + hash<symbol>()(e.type);
    }
  };
}     // namespace EXT_NAMESPACE

class eedge : public std::pair<int,eedge_label> { 
public:
  eedge(int pos_ = 0, symbol cat_ = symbol::undefined(), symbol type_ = symbol::undefined())
    : std::pair<int,eedge_label>(pos_, eedge_label(cat_, type_)) { }
};

typedef std::set<eedge> sE;

template <typename Tree>
class tree_eedges : public std::binary_function<const Tree*, sE, void> {

private:

  int tree_eedges_(const Tree* t, sE& es, int pos = 0) const
  {
    assert(t != NULL);
    if (t->label.cat == symbol::none)               // empty nodes
      return pos;

    if (t->label.cat.is_punctuation())              // punctuation preterminal
      return pos;

    if (t->child == NULL)                           // terminal
      return pos+1;

    bool empty = t->is_empty();
    if (empty) {                                    // empty non-terms
      symbol type = symbol::none;
      if (t->child 
	  && t->child->label.cat == symbol::none 
	  && t->child->next == NULL) {
	assert(t->child->child);
	assert(t->child->child->next == NULL);
	type = t->child->child->label.cat;
      }
      es.insert(eedge(pos, t->label.cat, type));
    }
  
    for (const Tree* c = t->child; c; c = c->next) {
      if (!empty && c->label.cat == symbol::none)  { // special case for empty complementizers
	assert(c->child);
	assert(c->child->next == NULL);
	es.insert(eedge(pos, symbol::none, c->child->label.cat));
      }
      pos = tree_eedges_(c, es, pos);
    }

    return pos;
  }

public:

  void operator() (const Tree* t, sE& es) const { tree_eedges_(t, es); }
};


// class tree_beedges{} is a function object mapping trees to a set of
//  `edges' for empty nodes that include their immediate antecedents
//
typedef std::pair<symbol,eedge_label> SL;
typedef std::pair<III,SL> IIISL;
typedef std::set<IIISL> sIIISL;

template <typename Tree>
class tree_beedges : public std::binary_function<const Tree*, sIIISL, void> 
{
private:

  typedef std::pair<II,symbol> IIS;
  typedef std::map<size_t,IIS> I_IIS;
  mutable const Tree* tree_root; // for debugging only

  int tree_binders(const Tree* t,  I_IIS& index_binder, int left = 0) const
  {
    assert(t);

    if (t->label.cat.is_punctuation() || t->label.cat == symbol::none) { 
      assert(t->child);           // punctuation or NONE preterminal
      assert(!t->child->next);
      return t->next ? tree_binders(t->next, index_binder, left) : left;
    }

    if (t->child == NULL) {       // terminal node
      assert(t->next == NULL);   
      return left+1;
    }

    int right = tree_binders(t->child, index_binder, left);
    if (t->label.index) {
      bool inserted = index_binder.insert(std::make_pair(t->label.index,
							 IIS(II(left, right), 
							     t->label.cat))
					  ).second;
      assert(inserted);
    }

    return t->next ? tree_binders(t->next, index_binder, right) : right;
  }

  int tree_empties(const Tree* t, const I_IIS& index_binder, sIIISL& beedges, int pos = 0) const
  {
    assert(t != NULL);
    if (t->label.cat == symbol::none)               // empty nodes
      return pos;

    if (t->label.cat.is_punctuation())              // punctuation preterminal
      return pos;

    if (t->child == NULL)                           // terminal
      return pos+1;

    bool empty = t->is_empty();
    if (empty) {                                    // empty non-terms
      symbol type = symbol::none;
      IIS antecedent(II(0, 0), symbol::none);
      if (t->child 
	  && t->child->label.cat == symbol::none
	  && t->child->next == NULL) {
	Tree* grandchild = t->child->child;
	assert(grandchild);
	assert(grandchild->next == NULL);
	type = grandchild->label.cat;
	if (grandchild->label.index) {
	  I_IIS::const_iterator it = index_binder.find(grandchild->label.index);
	  if (it != index_binder.end())
	    antecedent = it->second;
	  else 
	    std::cerr << "Error in empty.h tree_beedges::tree_empties() "
	      " Can't find an antecedent for " << t << " in " << std::endl
		      << tree_root << std::endl;
	}
      }
      beedges.insert(IIISL(III(antecedent.first, pos),
			   SL(antecedent.second, eedge_label(t->label.cat, type))));
    }
  
    for (const Tree* c = t->child; c; c = c->next) {
      if (!empty && c->label.cat == symbol::none)  { // special case for empty complementizers
	assert(c->child);
	assert(c->child->next == NULL);
	beedges.insert(IIISL(III(II(0,0), pos), 
			     SL(symbol::none,
				eedge_label(symbol::none, c->child->label.cat))));
      }
      pos = tree_empties(c, index_binder, beedges, pos);
    }

    return pos;
  }

public:

  void operator() (const Tree* t, sIIISL& beedges) const { 
    tree_root = t;
    I_IIS index_binder;
    tree_binders(t, index_binder);
    tree_empties(t, index_binder, beedges); 
  }
};
    
  

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            tree_hbeedges                           //
//                                                                    //
////////////////////////////////////////////////////////////////////////

// finds head of binder of empty node

template <typename Tree>
class tree_hbeedges : public std::binary_function<const Tree*, sIIISL, void> {

private:
  struct T_comp {
    bool operator()(const Tree* t1, const Tree* t2) const {
      return t1 < t2;
    }
  };

  typedef std::map<const Tree*,II,T_comp> T_II;

  size_t load_preterm_pos(const Tree* t, T_II& preterm_pos, size_t left = 0) const
  {
    if (t == NULL)
      return left;
    else if (t->is_nonterminal()) 
      return load_preterm_pos(t->next, preterm_pos, 
			      load_preterm_pos(t->child, preterm_pos, left));
    else {
      assert(t->is_preterminal());
      size_t right = 
	(t->label.cat.is_punctuation() || t->label.cat == symbol::none) ? left : left+1;
      bool inserted = preterm_pos.insert(std::make_pair(t, II(left,right))).second;
      assert(inserted);
      return load_preterm_pos(t->next, preterm_pos, right);
    }
  }

  typedef std::pair<II,symbol> IIS;
  typedef std::map<size_t,IIS> I_IIS;

  void tree_binder(const T_II& preterm_pos, const Tree* t,  I_IIS& index_binder) const
  {
    assert(t);

    if (t->next)
      tree_binder(preterm_pos, t->next, index_binder);

    if (!t->is_nonterminal())
      return;

    if (t->label.index) {
      const Tree* head = tree_syntacticLexicalHead(t);
      assert(head);
      assert(head->is_preterminal());
      bool inserted = index_binder.insert(std::make_pair(t->label.index,
							 IIS(afind(preterm_pos, head),
							     t->label.cat))
					  ).second;
      assert(inserted);
    }

    tree_binder(preterm_pos, t->child, index_binder);
  }

  int tree_empties(const Tree* t, const I_IIS& index_binder, sIIISL& hbeedges, int pos = 0) const
  {
    assert(t != NULL);
    if (t->label.cat == symbol::none)               // empty nodes
      return pos;

    if (t->label.cat.is_punctuation())              // punctuation preterminal
      return pos;

    if (t->child == NULL)                           // terminal
      return pos+1;

    bool empty = t->is_empty();
    if (empty) {                                    // empty non-terms
      symbol type = symbol::none;
      IIS antecedent(II(0,0), symbol::none);
      if (t->child && t->child->label.cat == symbol::none) {
	Tree* grandchild = t->child->child;
	assert(grandchild);
	assert(grandchild->next == NULL);
	type = grandchild->label.cat;
	if (grandchild->label.index) {
	  I_IIS::const_iterator it = index_binder.find(grandchild->label.index);
	  if (it != index_binder.end())
	    antecedent = it->second;
	}
      }
      hbeedges.insert(IIISL(III(antecedent.first, pos),
			    SL(antecedent.second, eedge_label(t->label.cat, type))));
    }
  
    for (const Tree* c = t->child; c; c = c->next) {
      if (!empty && c->label.cat == symbol::none)  { // special case for empty complementizers
	assert(c->child);
	assert(c->child->next == NULL);
	hbeedges.insert(IIISL(III(II(0,0), pos), 
			      SL(symbol::none,
				 eedge_label(symbol::none, c->child->label.cat))));
      }
      pos = tree_empties(c, index_binder, hbeedges, pos);
    }

    return pos;
  }

public:

  void operator() (const Tree* t, sIIISL& hbeedges) const { 
    T_II preterm_pos;
    load_preterm_pos(t, preterm_pos);
    I_IIS index_binder;
    tree_binder(preterm_pos, t, index_binder);
    tree_empties(t, index_binder, hbeedges); 
    // std::cout << "t = " << t << std::endl << "hbeedges = " << hbeedges << std::endl << std::endl;
  }
};


    
typedef std::map<eedge_label,III> EL_III;
  
#endif // EMPTY_H
