// heads.cc
//
// Mark Johnson, 9th November 2001

#include <cstdlib>
#include <ext/hash_map>
#include <vector>

#include "heads.h"
#include "sym.h"
#include "symset.h"
#include "utility.h"

struct Syntactic_head_data {
  typedef std::vector<symset> symsets;
  typedef ext::hash_map<symbol,const symsets*> sym_symsets;

  symsets adjective, conjunction, interjection, noun, preposition, unknown, verb;
  sym_symsets head_type;
  symset punctuation, rightheaded_nominals;

  Syntactic_head_data() : adjective(2), conjunction(1), interjection(1), 
			  noun(5), preposition(2), unknown(0), verb(4),
			  punctuation("# '' , -LRB- -RRB- . : ? ``"),
			  rightheaded_nominals("NN NNS NNP NNPS $") {
    adjective[0].set("$ CD JJ JJR JJS RB RBR RBS WRB");
    adjective[1].set("ADJP ADVP");
    conjunction[0].set("CC");
    interjection[0].set("INTJ UH");
    noun[0].set("DT WDT WP$ WP PRP EX");
    noun[1].set("NN NNS");
    noun[2].set("$ NNP NNPS");
    noun[3].set("-NONE- QP NP NP$ WHNP");
    noun[4].set("CD IN JJ JJR JJS PDT POS RB PP");
    preposition[0].set("IN RP TO");
    preposition[1].set("PP");
    verb[0].set("AUX AUXG MD");
    verb[1].set("VB VBD VBG VBN VBP VBZ");
    verb[2].set("VP");
    verb[3].set("ADJP JJ S SINV SQ TO");

    head_type[symbol("ADJP")] = &adjective;
    head_type[symbol("ADVP")] = &verb;
    head_type[symbol("CONJP")] = &conjunction;
    head_type[symbol("FRAG")] = &noun;
    head_type[symbol("INTJ")] = &interjection;
    head_type[symbol("LST")] = &noun;
    head_type[symbol("NAC")] = &noun;
    head_type[symbol("NP")] = &noun;
    head_type[symbol("NX")] = &noun;
    head_type[symbol("PP")] = &preposition;
    head_type[symbol("PRN")] = &noun;
    head_type[symbol("PRT")] = &preposition;
    head_type[symbol("QP")] = &noun;
    head_type[symbol("ROOT")] = &verb;
    head_type[symbol("RRC")] = &verb;
    head_type[symbol("S")] = &verb;
    head_type[symbol("SBAR")] = &verb;
    head_type[symbol("SBARQ")] = &verb;
    head_type[symbol("SINV")] = &verb;
    head_type[symbol("SQ")] = &verb;
    head_type[symbol("S1")] = &verb;
    head_type[symbol("UCP")] = &adjective;
    head_type[symbol("VP")] = &verb;
    head_type[symbol("WHADJP")] = &adjective;
    head_type[symbol("WHADVP")] = &adjective;
    head_type[symbol("WHNP")] = &noun;
    head_type[symbol("WHPP")] = &preposition;
    head_type[symbol("X")] = &unknown;
  }

  template <class Tree>
  const Tree* headchild(const Tree* t) const {
    if (t == NULL || t->child == NULL || t->child->child == NULL)
      return NULL;

    if (t->child->label.cat == symbol::none && t->child->next == NULL)
      return t->child;

    const Tree* head = NULL;    // search for head
    const sym_symsets::const_iterator it = head_type.find(t->label.cat);
    if (it == head_type.end()) {
      std::cerr << "Tree_head_data::headchild Error: can't find entry for category " 
		<< t->label.cat << std::endl << t << std::endl;
      abort();
    }
    const symsets& type = *it->second;
    cforeach (symsets, cats, type) {
      for (const Tree* child = t->child; child; child = child->next)
	if (cats->contains(child->label.cat)) {
	  head = child;
	  if (type == verb || type == preposition 
	      || (type == noun && !rightheaded_nominals.contains(child->label.cat)) )
	    break;
	}
      if (head != NULL)
	return head;
    }

    // didn't find a head; return right-most non-punctuation preterminal
    for (const Tree* child = t->child; child; child = child->next)
      if (child->is_preterminal() && !punctuation.contains(child->label.cat))
	head = child;
    if (head != NULL)
      return head;

    // still no head -- return right-most non-punctuation
    for (const Tree* child = t->child; child; child = child->next)
      if (!punctuation.contains(child->label.cat))
	head = child;
    if (head != NULL)
      return head;

    return NULL;
  }
};

const tree* tree_syntacticHeadChild(const tree* t)
{
  static Syntactic_head_data syntactic_head_data;

  return syntactic_head_data.headchild(t);
}

const tree* tree_syntacticLexicalHead(const tree* t)
{
  const tree* head;
  while ((head = tree_syntacticHeadChild(t)) != NULL)
    t = head;
  return t;
}


struct Semantic_head_data {
  typedef std::vector<symset> symsets;
  typedef ext::hash_map<symbol,const symsets*> sym_symsets;

  symsets adjective, conjunction, interjection, noun, preposition, unknown, verb;
  sym_symsets head_type;
  symset punctuation;

  Semantic_head_data() : adjective(2), conjunction(1), interjection(1), 
		     noun(4), preposition(2), unknown(0), verb(4),
		     punctuation("# '' , -LRB- -RRB- . : ? ``") {
    adjective[0].set("$ CD JJ JJR JJS RB RBR RBS WRB");
    adjective[1].set("ADJP ADVP");
    conjunction[0].set("CC");
    interjection[0].set("INTJ UH");
    noun[0].set("EX NN NNS PRP WP");
    noun[1].set("$ NNP NNPS");
    noun[2].set("QP NP WP$");
    noun[3].set("CD DT IN JJ JJR JJS PDT POS RB WDT");
    preposition[0].set("IN RP TO");
    preposition[1].set("PP");
    verb[0].set("VP");
    verb[1].set("VB VBD VBG VBN VBP VBZ");
    verb[2].set("ADJP JJ S SINV SQ TO");
    verb[3].set("AUX AUXG MD");

    head_type[symbol("ADJP")] = &adjective;
    head_type[symbol("ADVP")] = &verb;
    head_type[symbol("CONJP")] = &conjunction;
    head_type[symbol("FRAG")] = &noun;
    head_type[symbol("INTJ")] = &interjection;
    head_type[symbol("LST")] = &noun;
    head_type[symbol("NAC")] = &noun;
    head_type[symbol("NP")] = &noun;
    head_type[symbol("NX")] = &noun;
    head_type[symbol("PP")] = &preposition;
    head_type[symbol("PRN")] = &noun;
    head_type[symbol("PRT")] = &preposition;
    head_type[symbol("QP")] = &noun;
    head_type[symbol("ROOT")] = &verb;
    head_type[symbol("RRC")] = &verb;
    head_type[symbol("S")] = &verb;
    head_type[symbol("SBAR")] = &verb;
    head_type[symbol("SBARQ")] = &verb;
    head_type[symbol("SINV")] = &verb;
    head_type[symbol("SQ")] = &verb;
    head_type[symbol("S1")] = &verb;
    head_type[symbol("UCP")] = &adjective;
    head_type[symbol("VP")] = &verb;
    head_type[symbol("WHADJP")] = &adjective;
    head_type[symbol("WHADVP")] = &adjective;
    head_type[symbol("WHNP")] = &noun;
    head_type[symbol("WHPP")] = &preposition;
    head_type[symbol("X")] = &unknown;
  }

  template <class Tree>
  const Tree* headchild(const Tree* t) const {
    if (t == NULL || t->child == NULL || t->child->child == NULL)
      return NULL;

    if (t->child->label.cat == symbol::none && t->child->next == NULL)
      return t->child;

    const Tree* head = NULL;    // search for head
    const sym_symsets::const_iterator it = head_type.find(t->label.cat);
    if (it == head_type.end()) {
      std::cerr << "Tree_head_data::headchild Error: can't find entry for category " 
		<< t->label.cat << std::endl << t << std::endl;
      abort();
    }
    const symsets& type = *it->second;
    cforeach (symsets, cats, type) {
      for (const Tree* child = t->child; child; child = child->next)
	if (cats->contains(child->label.cat)) {
	  head = child;
	  if (type == verb || type == preposition)
	    break;
	}
      if (head != NULL)
	return head;
    }

    // didn't find a head; return right-most non-punctuation preterminal
    for (const Tree* child = t->child; child; child = child->next)
      if (child->is_preterminal() && !punctuation.contains(child->label.cat))
	head = child;
    if (head != NULL)
      return head;

    // still no head -- return right-most non-punctuation
    for (const Tree* child = t->child; child; child = child->next)
      if (!punctuation.contains(child->label.cat))
	head = child;
    if (head != NULL)
      return head;

    return NULL;
  }
};

const tree* tree_semanticHeadChild(const tree* t)
{
  static Semantic_head_data semantic_head_data;

  return semantic_head_data.headchild(t);
}

const tree* tree_semanticLexicalHead(const tree* t)
{
  const tree* head;
  while ((head = tree_semanticHeadChild(t)) != NULL)
    t = head;
  return t;
}
