// parseval.h
//
// Mark Johnson, 25th October 2001
//
// Write out precision/recall/f-score results
//
// report-ratio-threshold is an argument to write_parsevals
//  that thresholds the parseval data so that only high-count
//  results are printed

#ifndef PARSEVAL_H
#define PARSEVAL_H

#include <algorithm>
#include <functional>
#include <iostream>
#include <vector>

// parseval is a structure that exists simply for writing 
//  parseval results onto an output stream
//
struct parseval {
  size_t ncommon;
  size_t ntest;
  size_t ngold;
  parseval(size_t ncommon, size_t ntest, size_t ngold) 
    : ncommon(ncommon), ntest(ntest), ngold(ngold) { }
  parseval() : ncommon(0), ntest(0), ngold(0) { }
};

inline bool operator< (const parseval& p1, const parseval& p2) {
  return p1.ngold > p2.ngold;
}

inline std::ostream& operator<< (std::ostream& os, parseval s) {
  float prec = (s.ntest == 0 ? 0 : float(s.ncommon)/float(s.ntest));
  float rec = (s.ngold == 0 ? 0 : float(s.ncommon)/float(s.ngold));
  float f = (s.ncommon == 0 ? 0 : 2.0*prec*rec/(rec+prec));
  os << "Precision = " << s.ncommon << '/' << s.ntest << " = " << prec
     << ", recall = " << s.ncommon << '/' << s.ngold << " = " << rec
     << ", f-score = " << f;
  return os;
}

struct compare_second {
  template <typename T>
  bool operator() (const T& p1, const T& p2) const {
    return p1.second < p2.second;
  }
};


// write_parsevals() -- write out a map from labels to parseval
//
// report_ratio_threshold -- skip parsevals with less than this fraction
//   of total counts
//
template <typename Key_Parseval>
std::ostream& write_parsevals(const Key_Parseval& kps, std::ostream& os, 
			      float report_ratio_threshold = 0.1, size_t indent = 0)  
{
  parseval total;
  typedef typename Key_Parseval::const_iterator CI;
  for (CI it = kps.begin(); it != kps.end(); ++it) {
    total.ncommon += it->second.ncommon;
    total.ntest += it->second.ntest;
    total.ngold += it->second.ngold;
  }
  
  for (size_t i = 0; i < indent; ++i)
    os << ' ';

  os << "Overall results: " << total << std::endl << std::endl;

  typedef std::pair<typename Key_Parseval::key_type,typename Key_Parseval::data_type> P;
  typedef std::vector<P> Ps;
  Ps kps1;
  std::copy(kps.begin(), kps.end(), inserter(kps1, kps1.end()));
  std::sort(kps1.begin(), kps1.end(), compare_second());

  float report_count_threshold = report_ratio_threshold * total.ngold;

  for (typename Ps::const_iterator it = kps1.begin(); it != kps1.end(); ++it) 
    if (it->second.ngold >= report_count_threshold 
	|| it->second.ntest >= report_count_threshold) {
      for (size_t i = 0; i < indent; ++i)
	os << ' ';
      os << it->first << '\t' << it->second << std::endl;
    }
  return os;
}

#endif // PARSEVAL_H
