// pattern.cc
//
// Mark Johnson, 5th September, 2001

#include "pattern.h"
#include "utility.h"

#include <algorithm>
#include <cassert>
#include <iostream>
#include <map>

static const bool ignore_adjunction = true;

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                        tree patterns                               //
//                                                                    //
////////////////////////////////////////////////////////////////////////

std::ostream& operator<< (std::ostream& os, const pattern_tree_label& pl)
{
  os << tree_label(pl);
  
  switch (pl.type) {
  case normal:
    break;
  case subtree:
    os << "^subtree";
    break;
  case empty:
    os << "^empty";
    break;
  case adjunction:
    os << "^adjunction";
    break;
  case skeleton:
    os << "^skeleton";
    break;
  default:
    std::cerr << "Error in operator<<: unknown pattern_tree_label type " << pl.type << std::endl;
    exit(EXIT_FAILURE);
    break;
  }
  return os;
}

    
////////////////////////////////////////////////////////////////////////
//                                                                    //
//                           pattern_match()                          //
//                                                                    //
////////////////////////////////////////////////////////////////////////  

// pattern_substitute
//

template <bool copy_next>  
static tree* pattern_rcopy(const pattern_tree* pt, size_t index_base, size_t& index_max)
{
  if (pt == NULL)
    return NULL;
  size_t index = pt->label.index != 0 ? pt->label.index + index_base : 0;
  index_max = std::max(index, index_max);
  return new tree(pt->label.cat, index, 
		  pattern_rcopy<true>(pt->child, index_base, index_max),
		  copy_next ? pattern_rcopy<true>(pt->next, index_base, index_max) : NULL);
}

// pattern_substitute destructively changes t by inserting the empty nodes
//  in pt.  index_base is a number added to indices of all copied nodes
//
//  The pattern pt *must* match the tree t (use pattern_match to check this)
//  
//  pt == NULL indicates that the pattern does not apply to this part of the
//   tree
//
tree* pattern_substitute(const pattern_tree* pt, tree* t, size_t index_base, size_t& index_max)
{
  // NULL pattern -- doesn't change anything
  //
  if (pt == NULL) 
    return t;  // assume that pattern doesn't require us to match t

  // Empty pattern -- create new node and copy empty pattern subtree
  //
  if (pt->label.type == empty) {
    tree* t1 = pattern_rcopy<false>(pt, index_base, index_max);
    assert(t1 != NULL);
    assert(t1->next == NULL);
    t1->next = pattern_substitute(pt->next, t, index_base, index_max);
    return t1;
  }

  assert(t != NULL);

  // Skeleton pattern -- find matching node, change index if required
  //
  if (pt->label.type == skeleton) {
    if (pt->label.cat == t->label.cat) {
      if (pt->label.index != 0) {
	t->label.index = pt->label.index+index_base;
	index_max = std::max(index_max, t->label.index);
      }
      t->child = pattern_substitute(pt->child, t->child, index_base, index_max);
      if (pt->next)
	t->next = pattern_substitute(pt->next, t->next, index_base, index_max);
      return t;
    }
    else { // category match failed
      t->next = pattern_substitute(pt, t->next, index_base, index_max);
      return t;
    }
  }

  // Subtree pattern -- only change index
  //
  if (pt->label.type == subtree) {
    assert(pt->label.cat == t->label.cat);
    if (pt->label.index != 0) {
      t->label.index = pt->label.index+index_base;
      index_max = std::max(index_max, t->label.index);
    }
    t->next = pattern_substitute(pt->next, t->next, index_base, index_max);
    return t;
  }

  // Adjunction -- *t will become lower node in adjunction
  //
  if (pt->label.type == adjunction) {
    assert(pt->label.cat == t->label.cat);
    size_t index = pt->label.index != 0 ? pt->label.index+index_base : 0;
    index_max = std::max(index_max, index);
    tree* t1 = new tree(t->label.cat, index,
			NULL,
			pattern_substitute(pt->next, t->next, index_base, index_max));
    t->next = NULL;
    t1->child = pattern_substitute(pt->child, t, index_base, index_max);
    return t1;
  }

  // Normal tree
  //
  assert(pt->label.type == normal);
  assert(pt->label.cat == t->label.cat);
  if (pt->label.index != 0) {
    t->label.index = pt->label.index+index_base;
    index_max = std::max(index_max, t->label.index);
  }
  t->child = pattern_substitute(pt->child, t->child, index_base, index_max);
  t->next = pattern_substitute(pt->next, t->next, index_base, index_max);
  return t;
}
    



////////////////////////////////////////////////////////////////////////
//                                                                    //
//                        tree pattern index                          //
//                                                                    //
////////////////////////////////////////////////////////////////////////

// This is a specialized version for pattern_tree nodes that uses the type fields
//  on those nodes
//
template <>
void tree_pattern_index<pattern_tree_label>(const tree_node<pattern_tree_label>* pt, pattern_index& pi)
{
  pi.clear();
  pi.push_back(pt->label.cat);

  for (const pattern_tree* c = pt->child; c; c = c->next) {
    assert(c->label.type != skeleton);
    if (c->label.type != empty)
      pi.push_back(c->label.cat);
  }
}
