// pattern.h
//
// Mark Johnson, 4th September, 2001
//
//   class pattern_tree_label
//   pattern_tree
//   unindexed_empty_pattern_tree

#ifndef PATTERN_H
#define PATTERN_H

#include <vector>

#include "empty.h"
#include "tree.h"
#include "utility.h"

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                        tree patterns                               //
//                                                                    //
////////////////////////////////////////////////////////////////////////

enum node_type { normal = 0, subtree, empty, adjunction, skeleton };

class pattern_tree_label : public tree_label {
public:
  node_type type;

  pattern_tree_label(const tree_label& tree_label_) : tree_label(tree_label_), type(normal) { }

  pattern_tree_label(const symbol cat_ = symbol::undefined(), size_t index_ = 0, 
		     node_type type_ = normal)
    : tree_label(cat_, index_), type(type_) { };
};

std::ostream& operator<< (std::ostream& os, const pattern_tree_label& pl);

namespace std {   // these are needed for map and hash
  template <> struct equal_to<pattern_tree_label> 
    : public binary_function<pattern_tree_label,pattern_tree_label,bool> {
    bool operator() (const pattern_tree_label& l1, const pattern_tree_label& l2) const {
      return equal_to<tree_label>()(l1, l2) && l1.type == l2.type;
    }
  };

  template <> struct less<pattern_tree_label>
    : public binary_function<pattern_tree_label,pattern_tree_label,bool> {
    bool operator() (const pattern_tree_label& l1, const pattern_tree_label& l2) const {
      return less<tree_label>()(l1, l2) || (!less<tree_label>()(l2,l1) && l1.type < l2.type);
    }
  };
};  // namespace std

namespace EXT_NAMESPACE {
  template <> struct hash<pattern_tree_label> {
    size_t operator()(const pattern_tree_label& l) const {
      return hash<tree_label>()(l)+5*l.type;
    }
  };
}; // namespace EXT_NAMESPACE

typedef tree_node<pattern_tree_label> pattern_tree;

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                           pattern_match()                          //
//                                                                    //
////////////////////////////////////////////////////////////////////////

template <typename tree_label>
static bool pattern_match_adjunction(const pattern_tree* pt, const tree_node<tree_label>* t);

// pattern_match returns t if pt matches a tree t without empty nodes
//
// This auxiliary function handles the recursive case
//
template <typename tree_label>
static bool pattern_match_(const pattern_tree* pt, const tree_node<tree_label>* t)
{
  if (pt == NULL)
    return t == NULL;

  if (pt->label.type == empty) 
    return pattern_match_(pt->next, t);
  
  if (t == NULL)
    return false;

  if (pt->label.type == skeleton) {
    if (pt->label.cat == t->label.cat)
      return (pt->label.index == 0 || t->label.index == 0) 
	     && pattern_match_(pt->child, t->child) 
	     && (pt->next == NULL || pattern_match_(pt->next, t->next));
    else return pattern_match_(pt, t->next);  // cat failed to match; try next tree
  }

  if (pt->label.type == subtree)
    return pt->label.cat == t->label.cat 
           && (pt->label.index == 0 || t->label.index == 0)
           && pattern_match_(pt->next, t->next);

  if (pt->label.type == adjunction)
    return pattern_match_adjunction(pt, t) && pattern_match_(pt->next, t->next);

  assert(pt->label.type == normal);
  return pt->label.cat == t->label.cat 
         && (pt->label.index == 0 || t->label.index == 0)
         && pattern_match_(pt->child, t->child) 
         && pattern_match_(pt->next, t->next);
}

template <typename tree_label>
static bool pattern_match_adjunction(const pattern_tree* pt, const tree_node<tree_label>* t)
{
  assert(pt);
  if (pt->label.cat != t->label.cat 
      || (pt->label.index != 0 && t->label.index != 0))
    return false;
  const pattern_tree* cpt = pt->child;
  assert(cpt != NULL);
  while (cpt && cpt->label.type == empty)
    cpt = cpt->next;
  assert(cpt != NULL);
  assert(cpt->label.cat == pt->label.cat);
  bool match = true;
  if (cpt->label.type == normal)
    match = pattern_match_(cpt->child, t->child);
  for (cpt = cpt->next; cpt; cpt = cpt->next)
    assert(cpt->label.type == empty);
  return match;
}

template <typename tree_label>
bool pattern_match(const pattern_tree* pt, const tree_node<tree_label>* t)
{
  assert(pt != NULL);
  assert(t != NULL);

  if (pt->label.type == adjunction) 
    return pattern_match_adjunction(pt, t);

  // if (pt->label.type != normal) 
  //   std::cout << "Unexpected pattern tree type:\n" << pt << std::endl;
    
  assert(pt->label.type == normal);

  return pt->label.cat == t->label.cat 
        && (pt->label.index == 0 || t->label.index == 0)
        && pattern_match_(pt->child, t->child);
}


// pattern_substitute destructively changes t by inserting the empty nodes
//  in pt.  index_base is a number added to indices of all copied nodes.
//  
//  The pattern pt *must* match the tree t (use pattern_match to check this)
//  
tree* pattern_substitute(const pattern_tree* pt, tree* t, size_t index_base, size_t& index_max);


////////////////////////////////////////////////////////////////////////
//                                                                    //
//                        tree pattern index                          //
//                                                                    //
////////////////////////////////////////////////////////////////////////

typedef std::vector<symbol> pattern_index;

// This maps a tree into a vector of the top node's category and a list
//  of the categories of the non-empty nodes immediately below
//
template <typename label_type>
void tree_pattern_index(const tree_node<label_type>* t, pattern_index& pi)
{
  pi.clear();
  pi.push_back(t->label.cat);

  for (const tree_node<label_type>* c = t->child; c; c = c->next)
    if (!c->is_empty())
      pi.push_back(c->label.cat);
}
 
// This is a specialized version for pattern_tree nodes that uses the type fields
//  on those nodes
//
template <>
void tree_pattern_index<pattern_tree_label>(const tree_node<pattern_tree_label>* pt, pattern_index& pi);



#endif // PATTERN_H
