/* read-pattern-tree.l
 *
 * Mark Johnson, 17th July 2002
 */

%option noyywrap
%option fast
%option 8bit
%option prefix="read_pattern_tree"

%{
#include "sym.h"
#include "read-pattern-tree.h"

#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <stack>

const char* read_pattern_tree_filename = NULL;

inline void message(const char* s1, const char* s2) {
  fprintf(stderr, "%s: %s %s\n", read_pattern_tree_filename, s1, s2);
}

#define YY_DECL static pattern_tree* read_pattern_tree_lex()

%}

/* RT = root
 * FC = first child
 * NC = next child
 */

%s RT
%s FC
%s NC
%x SKIP_LINE

CAT   ([^ \t\n()^-]+|"-NONE-"|"-RRB-"|"-LRB-")

%%

%{
  pattern_tree* root = NULL;       /* tree's root node */
  std::stack<pattern_tree*> s;     /* stack of node pointers */

  #define NEW_CHILD(txt)   assert(!s.empty()); s.push(s.top()->child = new pattern_tree); s.top()->label.cat = symbol(txt);
  #define SWAP_NEXT(txt)   assert(!s.empty()); (s.top() = s.top()->next = new pattern_tree)->label.cat = symbol(txt);
%}

<RT>"("{CAT}    assert(s.empty()); s.push(root = new pattern_tree); s.top()->label.cat = symbol(yytext+1); BEGIN(FC);

<FC>"("{CAT}    NEW_CHILD(yytext+1); BEGIN(FC);
<FC>{CAT}       NEW_CHILD(yytext); BEGIN(NC);

<NC>"("{CAT}    SWAP_NEXT(yytext+1); BEGIN(FC);
<NC>{CAT}       SWAP_NEXT(yytext); BEGIN(NC);

<NC>")"         assert(!s.empty()); s.pop(); if (s.size() == 1) return root; BEGIN(NC);

"-"[0-9]+	{ assert(!s.empty()); 
                  assert(s.top()->label.index == 0); 
		  s.top()->label.index = strtoul(yytext+1, NULL, 10); 
		  assert(s.top()->label.index != 0); 
                }

"^subtree"	assert(!s.empty()); assert(s.top()->label.type == normal); s.top()->label.type = subtree;
"^empty"        assert(!s.empty()); assert(s.top()->label.type == normal); s.top()->label.type = empty;
"^adjunction"   assert(!s.empty()); assert(s.top()->label.type == normal); s.top()->label.type = adjunction;
"^skeleton"     assert(!s.empty()); assert(s.top()->label.type == normal); s.top()->label.type = skeleton;

[ \t\n]+	/* ignore spaces */
.		message("Unexpected character", yytext); std::cerr << "Parse tree so far: " << root << '\n'; abort();

<SKIP_LINE>[^ \t\n()]+   /* ignore words */
<SKIP_LINE>[ \t\n]+      /* ignore spaces */
<SKIP_LINE>"("           /* ignore open parentheses */
<SKIP_LINE>")"           return(NULL);   /* return something when you hit a parenthesis */

%%

pattern_tree* read_pattern_tree()
{
  BEGIN(RT);
  pattern_tree* pt = read_pattern_tree_lex();
  // std::cout << "pt = " << pt << std::endl;
  return pt;
}

void read_pattern_tree_skip_line()
{
  BEGIN(SKIP_LINE);
  read_pattern_tree_lex();
}
