// restore_empties.h
//
// Mark Johnson, 25th September 2001
//
// last modified 29th October 2001
//
// A modification of restore_empties_lex.h that selectively includes nodes
// in a pattern, controlled by the function object Select

#ifndef RESTORE_EMPTIES_H
#define RESTORE_EMPTIES_H

#include <algorithm>
#include <cassert>
#include <cstdio>
#include <ext/hash_map>
#include <iostream>
#include <iterator> 
#include <list>
#include <map>
#include <vector>

#include "shared_ptr.hpp"

#include "cbinom-approx.h"
#include "empty.h"
#include "pattern.h"
#include "read-pattern-tree.h"
#include "sym.h"
#include "tree.h"
#include "utility.h"


template <typename Select, bool skeletal = false, bool embedded_patterns = true>
class stree_lex_pattern {
private:

  struct relabeler : public std::map<size_t,size_t> {
    typedef tree_label::index_type index_type;

    index_type load(const index_type original_index) {
      if (original_index == 0)     // 0 is null index
	return 0;     
      index_type index = insert(std::make_pair(original_index,size()+1)).first->second;
      assert(index>0);
      return index;
    }

    index_type relabel(const index_type original_index) const {
      const_iterator it = find(original_index);
      if (it != end())
	return it->second;
      else
	return 0;      // return no index
    }

    bool disjoint(const sI& s) const {
      cforeach (sI, it, s)    // there ought to be a better way
	if (count(*it) > 0)   //  that doesn't involve blind lookup
	  return false;
      return true;
    }
  };

  relabeler    r;
  Select       select;
  const stree* root_node;

  // is_on_slash_path is true iff one of t's children has a slash index 
  //  that appears in the relabeller for the pattern being extracted
  //
  bool is_on_slash_path(const stree* t) const
  {
    if (t != NULL)
      for (const stree* c = t->child; c; c = c->next)
	if (!r.disjoint(c->label.slashes))
	  return true;
    return false;
  }

  pattern_tree* copy_subtree(const stree* t) const
  {
    if (t == NULL)
      return NULL;
    else
      return new pattern_tree(t->label.cat, r.relabel(t->label.index),
			      copy_subtree(t->child),
			      copy_subtree(t->next));
  }

  pattern_tree* copy_children(const stree* t, bool toplevel = false) const
  {
    if (t == NULL)
      return NULL;
    
    // if this is a deeper part of the pattern, don't copy empty nodes not
    //  on slash path
    //
    if (toplevel == false && t->is_empty() && r.disjoint(t->label.slashes) 
	&& disjoint(root_node->label.slashes, t->label.slashes))
      return copy_children(t->next);

    pattern_tree* pt = new pattern_tree(t->label.cat, r.relabel(t->label.index));
    
    if (t->is_empty() && disjoint(root_node->label.slashes, t->label.slashes)) {
      pt->label.type = empty;
      pt->child = copy_subtree(t->child);
    }
    else if (select(t)) {                        // copy t's children if select() tells us to
      pt->child = copy_children(t->child);
      if (pt->adjunction_site())
	pt->label.type = adjunction;
    }      
    else if (is_on_slash_path(t)) {              // does t have a slash seen at top-level?
      pt->child = copy_children(t->child);
      if (pt->adjunction_site())
	pt->label.type = adjunction;
    }
    else if (!t->is_terminal())
      pt->label.type = subtree;

    pt->next = copy_children(t->next, toplevel);

    return pt;
  }

  // skeletalize deletes all subtree nodes and converts all normal nodes
  //  to skeleton nodes.  It doesn't touch any local trees with adjunctions
  //  or empty nodes in them, though
  //
  pattern_tree* skeletalize_siblings(pattern_tree* pt) const
  {
    if (pt == NULL)
      return NULL;
    
    pt->next = skeletalize_siblings(pt->next);

    if (pt->label.type == normal) 
      pt->label.type = skeleton;     
    else if (pt->label.type == subtree) {    // decide whether to delete this node
      if (pt->label.index == 0 &&            // don't delete indexed nodes
	  (pt->next == NULL || pt->next->label.type == skeleton)) {  
	bool same_cat_follows = false;
	for (const pattern_tree* n = pt->next; n; n = n->next) 
	  if (n->label.cat == pt->label.cat) {
	    same_cat_follows = true;
	    break;
	  }
	if (same_cat_follows == false) { 
	  // delete node
	  pattern_tree* ptn = pt->next;
	  pt->next = NULL;
	  delete pt;
	  return ptn;
	}}
    }
    return pt;
  }
	
  void skeletalize_tree(pattern_tree* pt) const
  {
    if (pt == NULL)
      return;

    bool skel = true;
    for (pattern_tree* c = pt->child; c; c = c->next) {
      if (c->label.type != empty && c->label.type != adjunction)
	skeletalize_tree(c);
      if ((c->label.type != subtree && c->label.type != normal) || c->label.index != 0)
	skel = false;
    }

    if (skel == true) 
      pt->child = skeletalize_siblings(pt->child);
  }

  pattern_tree* skeletal_copy(const pattern_tree* pt0) const 
  {
    assert(pt0 != NULL);
    pattern_tree* pt = pt0->copy_tree();
    for (pattern_tree* c = pt->child; c; c = c->next)
      if (c->label.type == normal)
	skeletalize_tree(c);
    return pt;
  }

public:

  stree_lex_pattern() : select() { }

  // Extract a pattern_tree from this tree node
  //   (this code is not re-entrant, but that shouldn't
  //   be a problem)
  //
  pattern_tree* tree_basepattern(const stree* t)
  {
    assert(t);
    if (!t->child || t->label.cat == symbol::none)
      return NULL;
    
    root_node = t;   // save root node

    r.clear();
    int n_children(0), n_empty(0), n_empty_unindexed(0), n_same_cat(0);
    
    // count empty nodes in tree, load relabeller
    //
    for (const stree* c = t->child; c; c = c->next) {
      ++n_children;
      if (c->is_empty()) {
	++n_empty;
	if (c->label.slashes.empty())
	  ++n_empty_unindexed;
      }
      else if (c->label.cat == t->label.cat)
	++n_same_cat;
      cforeach (sI, it, c->label.slashes)
	if (t->label.slashes.count(*it) == 0)
	  r.load(*it);
    };

    // if (r.empty() && (n_empty == 0 || n_empty == n_children))  // Bug that actually improves things!
    if (r.empty() && (n_empty_unindexed == 0 || n_empty == n_children))
      return NULL;

    if (n_empty+1 == n_children && n_same_cat == 1) return NULL;  

    pattern_tree* pt = new pattern_tree(pattern_tree_label(t->label.cat, r.relabel(t->label.index)));

    if (n_empty+1 == n_children && n_same_cat == 1)
      pt->label.type = adjunction;
    
    pt->child = copy_children(t->child, true);

    if (!pt->fillers_equal_gaps()) {
      std::cerr << "In restore-empties.h stree_lex_pattern::tree_basepattern() fillers and gaps don't match" 
		<< std::endl << " pt = " << pt << std::endl
		<< " t = " << t << std::endl;
      abort();
    }

    return pt;
  } // tree_basepattern()

  template <typename Patterns>
  void tree_patterns(const stree* t, Patterns& pts)
  {
    if (t == NULL)
      return;

    if (pattern_tree* pt = tree_basepattern(t)) {
      pts.push_back(pt);
      if (skeletal)
	pts.push_back(skeletal_copy(pt));
    }
    else if (embedded_patterns) 
      for (const stree* c = t->child; c; c = c->next) 
	if (pattern_tree* pt = tree_basepattern(c)) {
	  pattern_tree* parentpt = new pattern_tree(t->label.cat);
	  pattern_tree** cb = &(parentpt->child);
	  for (const stree* c1 = t->child; c1; c1 = c1->next) {
	    if (c1 == c) 
	      *cb = pt;
	    else 
	      *cb = new pattern_tree(pattern_tree_label(c1->label.cat, 0, subtree));
	    cb = &((*cb)->next);
	  }
	  pts.push_back(parentpt);
	  if (skeletal)
	    pts.push_back(skeletal_copy(parentpt));
	}
  } // tree_patterns()
};

struct select_none {   // a function object that is true of no nodes
  bool operator() (const stree* t) const {
    return false;
  }
};

struct select_aux {   // a function object that is true of preterminal nodes
  bool operator() (const stree* t) const {
    return t->is_preterminal() && t->child->label.cat.is_auxiliary();
  }
};

struct select_preterms {   // a function object that is true of preterminal nodes
  bool operator() (const stree* t) const {
    return t->is_preterminal();
  }
};

struct select_vbn {   // a function object that is true of VBN
  bool operator() (const stree* t) const {
    static const symbol VBN("VBN");
    return t->is_preterminal() && t->label.cat == VBN;
  }
};

//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//                           class restore_empties                          //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////
 
template <class Stree_pattern, bool print_mismatches = false>
class restore_empties {
 
  typedef boost::shared_ptr<pattern_tree> PT;
  typedef ext::hash_map<PT, size_t> PT_I;
  typedef std::vector<PT> PTs;
  typedef ext::hash_map<pattern_index, PTs> PI_PTs;

  typedef float Float;
  Float prune_ratio;    // prune pattern if correct match count / match count is less than this value

  PI_PTs pi_pts;        // index -> { pattern }

  struct data {
    PT_I    pt_count;   // pattern -> # of times pattern occured in training data
    PI_PTs& pi_pts;     // index -> { pattern }
    PT_I    pt_mcount;  // pattern -> # of times pattern matches some subtree in training data

    // The rest of these variables are for debugging purposes only
    size_t n_pass1_mismatches;
    size_t n_pass2_mismatches;
    const stree* current_tree;                    // to print out in pass2a
    std::auto_ptr<pattern_tree> current_pattern;  // to print out in pass2a

    data(PI_PTs& pi_pts_) : pi_pts(pi_pts_), n_pass1_mismatches(0), 
			    n_pass2_mismatches(0), current_tree(NULL) { }

    struct data1 {
      data& d;
    
      data1(data& d) : d(d) { }

      tree* copy_without_empties(const stree* t, const sI& root_slashes, tree* endp = NULL) const {
	if (t->is_empty() && disjoint(t->label.slashes, root_slashes))
	  return t->next ? copy_without_empties(t->next, root_slashes, endp) : endp;
	else {
	  tree* c = new tree(t->label.cat,
			     t->child ? copy_without_empties(t->child, root_slashes) : NULL,
			     t->next ? copy_without_empties(t->next, root_slashes, endp) : endp);
	  c->label.index = 0;
	  return c;
	}
      } // data1::copy_without_empties()
    
      void operator() (const stree* t) {
	typedef std::list<pattern_tree*> Patterns;
	Patterns pts;
	Stree_pattern().tree_patterns(t, pts);
	cforeach (Patterns, pti, pts) {
	  PT pt(*pti);
	  // std::cerr << t << std::endl << pti << std::endl;
	  ++d.pt_count[pt];
	  const tree* t1 = copy_without_empties(t, t->label.slashes);
	  if (!pattern_match(*pti, t1)) {
	    ++d.n_pass1_mismatches;
	    // if (print_mismatches)
	      std::cerr << "Error in pass1() of restore_empties: pattern *pt \n " << *pti << std::endl
			<< " does not match source subtree t1\n " << t1 << std::endl
			<< " original tree t\n " << t << std::endl;
	  }
	  delete t1;
	}
      } // data1::operator()
    };  // struct data1

    void pass1(const tree* t) {
      std::auto_ptr<stree> s(tree_stree(t));
      data1 d1(*this);
      // s->preorder(d1);   // is this correct?
      s->postorder(d1);     // no, but it doesn't make any difference
    } // pass1()

    void index_patterns() {
      pi_pts.clear();
      pattern_index ind;
      cforeach (PT_I, pi, pt_count) {
	tree_pattern_index(pi->first.get(), ind);
	pi_pts[ind].push_back(pi->first);
      }
    } // index_patterns()

    // pass2a counts how often each pattern matches a tree with
    //  all empty nodes and indices removed from it.
    //
    tree* pass2a_(tree* t, const stree* st, size_t& index_base) {

      // check that the tree t and the stree st are structurally isomorphic
      //
      if (t == NULL) {
	if (st != NULL) {
	  ++n_pass2_mismatches;
	  if (print_mismatches)
	    std::cerr << "Error in restore_empties::pass2a_: t = NULL, st = " << st << std::endl
		      << " current_tree = " << current_tree << std::endl
		      << " current_pattern = " << current_pattern << std::endl;
	  return t;
	}
	return t;
      }
      assert(st != NULL);
      if (t->label.cat != st->label.cat) {
	++n_pass2_mismatches;
	if (print_mismatches)
	  std::cerr << "Error in restore_empties::pass2a: Trees differ in structure\n"
		    << " t = " << t << std::endl
		    << " st = " << st << std::endl
		    << " current_tree = " << current_tree << std::endl
		    << " current_pattern = " << current_pattern << std::endl;
	return t;
      }

      // count the number of times each pattern matches this local tree
      //
      pattern_index ind;
      tree_pattern_index(t, ind);
      PI_PTs::const_iterator pi_pts_it = pi_pts.find(ind);
      if (pi_pts_it != pi_pts.end())
	cforeach (PTs, pt_it, pi_pts_it->second)
	  if (pattern_match(pt_it->get(), t))
	    ++pt_mcount[*pt_it];
      
      // now restore any empty nodes that ought to be restored here
      //
      pattern_tree* pt = Stree_pattern().tree_basepattern(st);
      if (pt != NULL) {
	if (pattern_match(pt, t))
	  t = pattern_substitute(pt, t, index_base, index_base);
	else {
	  ++n_pass2_mismatches;
	  if (print_mismatches)
	    std::cerr << "Error in restore_empties::pass2a: Extracted pattern does not match tree\n"
		      << " t = " << t << std::endl
		      << " st = " << st << std::endl
		      << " pt = " << pt << std::endl;
	  // abort();
	}
	current_pattern.reset(pt);
	// delete pt;
      }
      
      t->child = pass2a_(t->child, st->child, index_base);
      t->next = pass2a_(t->next, st->next, index_base);
      return t;
    }  // pass2a_()

    void pass2a(const tree* t0) {
      tree* t = t0->copy_without_empties();
      std::auto_ptr<stree> st(tree_stree(t0));
      current_tree = st.get();
      size_t index_base(0);
      t = pass2a_(t, st.get(), index_base);
      delete t;
    } // pass2a()
      

    Float success_prob(const PT& pt) const {
      size_t n = afind(pt_mcount, pt);
      size_t k = afind(pt_count, pt);
      return (n >= k) ? Float(k)/Float(n) : 1.0;
    }

    Float discounted_success_prob(const PT& pt) const {
      size_t n = afind(pt_mcount, pt);
      size_t k = afind(pt_count, pt);
      return cbinom_lower_p((n >= k) ? n : k, k);
    }

    std::ostream& prettyprint(std::ostream& os) const {
      cforeach (PI_PTs, pi_pt_it, pi_pts) {
	os << pi_pt_it->first << ':' << std::endl;
	cforeach (PTs, pt_it, pi_pt_it->second) {
	  size_t pattern_match_count = dfind(pt_count, *pt_it);
	  size_t pattern_count = dfind(pt_mcount, *pt_it);
	  os << " (" << pattern_match_count << "/" << pattern_count 
	     << ") = " << double(pattern_match_count)/double(pattern_count) 
	     << '\t' << *pt_it << std::endl;
	}
      }
      return os;
    }

    template <typename Ordering>
    std::ostream& prettyprint_ranked(std::ostream& os, const Ordering ordering) const {
      int n_to_print = 100;
      PTs pts;
      cforeach (PI_PTs, pi_pt_it, pi_pts)
	cforeach (PTs, pt_it, pi_pt_it->second)
 	  pts.push_back(*pt_it);
      std::sort(pts.begin(), pts.end(),	pattern_compare_count(*this));

      PTs::iterator end = pts.end();
      if (end - pts.begin() > n_to_print)
	end = pts.begin() + n_to_print;
      std::sort(pts.begin(), end, ordering);

      for (PTs::const_iterator pi = pts.begin(); pi < end; ++pi) 
	// os << "(" << dfind(pt_count, *pi) << "/" << dfind(pt_mcount, *pi)
	//    << ") = " << discounted_success_prob(*pi) << "\t" << *pi << std::endl;
	os << dfind(pt_count, *pi) << " & " << dfind(pt_mcount, *pi) << " & " << *pi << std::endl;
      return os;
    }

  };  // struct data

  struct pattern_compare_max_depth {
    const data& d;
    pattern_compare_max_depth(const data& d) : d(d) { }
    bool operator() (const PT& p1, const PT& p2) const {
      size_t d1 = p1->max_depth();
      size_t d2 = p2->max_depth();
      return d1 > d2 || (d1 == d2 && d.success_prob(p1) > d.success_prob(p2));
    }
  };

  struct pattern_compare_size_prob {
    const data& d;
    pattern_compare_size_prob(const data& d) : d(d) { }
    bool operator() (const PT& p1, const PT& p2) const {
      size_t s1 = p1->size();
      size_t s2 = p2->size();
      return s1 > s2 || (s1 == s2 && d.success_prob(p1) > d.success_prob(p2));
    }
  };

  struct pattern_compare_prob {
    const data& d;
    pattern_compare_prob(const data& d) : d(d) { }
    bool operator() (const PT& p1, const PT& p2) const {
      return d.success_prob(p1) > d.success_prob(p2);
    }
  };

  struct pattern_compare_discounted_prob {
    const data& d;
    pattern_compare_discounted_prob(const data& d) : d(d) { }
    bool operator() (const PT& p1, const PT& p2) const {
      return d.discounted_success_prob(p1) > d.discounted_success_prob(p2);
    }
  };

  struct pattern_compare_count {
    const data& d;
    pattern_compare_count(const data& d) : d(d) { }
    bool operator() (const PT& p1, const PT& p2) const {
      return afind(d.pt_count, p1) > afind(d.pt_count, p2);
    }
  };

public:

  restore_empties(Float prune_ratio_ = 0.5) : prune_ratio(prune_ratio_) { }

  template<typename Trees>
  restore_empties(const Trees& trees, Float prune_ratio_ = 0.5) 
    : prune_ratio(prune_ratio_)
  {
    typedef typename Trees::value_type TreePtr;
    typedef typename std::iterator_traits<TreePtr>::value_type Tree;

    data d(pi_pts);

    cforeach (typename Trees, ti, trees)
      d.pass1(*ti);

    // std::cerr << "   Before erasing useless entries d.pt_count.size() = " << d.pt_count.size() << std::endl;

    d.index_patterns();

    cforeach (typename Trees, ti, trees)
      d.pass2a(*ti);
    
    if (d.n_pass1_mismatches > 0)
      std::cerr << "   Note in restore_empties: number of patterns that failed to match n_pass1_mismatches = "
		<< d.n_pass1_mismatches << std::endl;

    if (d.n_pass2_mismatches > 0)
      std::cerr << "   Note in restore_empties: number of patterns that failed to match n_pass2_mismatches = "
		<< d.n_pass2_mismatches << std::endl;

    if (d.pt_count.size() != d.pt_mcount.size())
      std::cerr << "   Note in restore_empties: d.pt_count.size() = " << d.pt_count.size()
		<< ", d.pt_mcount.size() = " << d.pt_mcount.size() << std::endl;
		
    {
      PTs to_erase;
      cforeach (PT_I, pi, d.pt_count) {
	size_t n = dfind(d.pt_mcount, pi->first);
	size_t k = pi->second;
	if (n == 0 || cbinom_lower_p(std::max(n,k), k) < prune_ratio)
	  to_erase.push_back(pi->first);
      }

      cforeach (PTs, tei, to_erase) {
	d.pt_count.erase(*tei);
	d.pt_mcount.erase(*tei);
      }
    }
      
    // std::cerr << "   After erasing useless entries d.pt_count.size() = " << d.pt_count.size() << std::endl;

    d.index_patterns();

    // sort the patterns in each index
    //
    pattern_compare_max_depth ordering(d);
    // pattern_compare_prob ordering(d);
    foreach (PI_PTs, pi_pts_it, pi_pts)
      std::sort(pi_pts_it->second.begin(), pi_pts_it->second.end(), ordering);

    // d.prettyprint(std::cerr);
    // d.prettyprint_ranked(std::cerr, ordering);
  }


private:

  template <typename Tree>
  Tree* restore_empty_node(Tree* t, size_t& index_base) const {
    assert(t != NULL);
    pattern_index  ind;
    tree_pattern_index(t, ind);
    PI_PTs::const_iterator pi_pts_it = pi_pts.find(ind);
    if (pi_pts_it != pi_pts.end())
      cforeach (PTs, pt_it, pi_pts_it->second)
	if (pattern_match(pt_it->get(), t))
	  return pattern_substitute(pt_it->get(), t, index_base, index_base);
    return t;
  }

  template <typename Tree>
  Tree* restore_empty_nodes(Tree* t, size_t& index_base) const {
    if (t == NULL)
      return NULL;
    t = restore_empty_node(t, index_base);
    t->child = restore_empty_nodes(t->child, index_base);
    t->next = restore_empty_nodes(t->next, index_base);

    return t;
  }

public:

  template <typename Tree>
  Tree* operator() (Tree* t) const {
    size_t index_base = 0;
    return restore_empty_nodes(t, index_base);
  }
  
  std::ostream& write(std::ostream& os) const {
    cforeach (PI_PTs, pi_pt_it, pi_pts) {
      // os << pi_pt_it->first << ':' << std::endl; // print index
      cforeach (PTs, pt_it, pi_pt_it->second) 
	os << *pt_it << std::endl;
    }
    return os;
  }

  size_t read(FILE* fp) {
    read_pattern_treein = fp;
    read_pattern_tree_skip_line();
    pi_pts.clear();
    pattern_tree* pt;
    size_t nread = 0;
    while ((pt = read_pattern_tree())) {
      ++nread;
      pattern_index ind;
      tree_pattern_index(pt, ind);
      pi_pts[ind].push_back(PT(pt));
    }
    return nread;
  }

}; // class restore_empties

template <class Stree_pattern>
std::ostream& operator<< (std::ostream& os, const restore_empties<Stree_pattern>& re)
{
  return re.write(os);
}

#endif // RESTORE_EMPTIES_H
