// sym.cc
//
// (c) Mark Johnson, 10th March 2001
// (c) Mark Johnson, 12th December 2001 (fix static initialization order bug)


#include "sym.h"
#include <cctype>

#define ESCAPE     '\\'
#define OPENQUOTE  '\''
#define CLOSEQUOTE '\''
#define UNDEFINED  "%UNDEFINED%"           // UNDEFINED must start with punctuation

// define these as globals to avoid static initialization order bugs
//
symbol::Table& symbol::table() 
{
  static Table table_;
  return table_;
}

symbol::symbol(const std::string& s) : sp(&*(table().insert(s).first)) { };

symbol::symbol(const char* cp) { 
  if (cp) {
    std::string s(cp); 
    sp = &*(table().insert(s).first);
  }
  else
    sp = NULL;
};

// predefine some standard symbols
//
const symbol symbol::root("S1");     // make root label match Eugene's trees
const symbol symbol::none("-NONE-");
const symbol symbol::_("_");
const symbol symbol::dollar("$");
const symbol symbol::CD("CD");
const symbol symbol::NP("NP");
const symbol symbol::S("S");
const symbol symbol::SBAR("SBAR");
const symbol symbol::SINV("SINV");
const symbol symbol::VP("VP");

// punctuation

const symbol symbol::comma(",");
const symbol symbol::period(".");
const symbol symbol::forward_quote("``");
const symbol symbol::backward_quote("''");
const symbol symbol::colon(":");
const symbol symbol::lrb("-LRB-");
const symbol symbol::rrb("-RRB-");

const symbol symbol::be("be");
const symbol symbol::am("am");
const symbol symbol::are("are");
const symbol symbol::is("is");
const symbol symbol::was("was");
const symbol symbol::were("were");
const symbol symbol::being("being");
const symbol symbol::been("been");
const symbol symbol::apostrophe_m("'m");

const symbol symbol::have("have");
const symbol symbol::has("has");
const symbol symbol::had("had");
const symbol symbol::having("having");
const symbol symbol::apostrophe_ve("'ve");
const symbol symbol::apostrophe_d("'d");
  
const symbol symbol::will("will");
const symbol symbol::would("would");
const symbol symbol::can("can");
const symbol symbol::could("could");
const symbol symbol::shall("shall");
const symbol symbol::should("should");

const symbol symbol::Be("Be");
const symbol symbol::Am("Am");
const symbol symbol::Are("Are");
const symbol symbol::Is("Is");
const symbol symbol::Was("Was");
const symbol symbol::Were("Were");
const symbol symbol::Being("Being");
const symbol symbol::Been("Been");

const symbol symbol::Have("Have");
const symbol symbol::Has("Has");
const symbol symbol::Had("Had");
const symbol symbol::Having("Having");
  
const symbol symbol::Will("Will");
const symbol symbol::Would("Would");
const symbol symbol::Can("Can");
const symbol symbol::Could("Could");
const symbol symbol::Shall("Shall");
const symbol symbol::Should("Should");

symbol symbol::to_lower() const {
  const std::string& s = string_reference();
  std::string l;
  for (std::string::const_iterator si = s.begin(); si != s.end(); ++si)
    if (isupper(*si))
      l.push_back(tolower(*si));
    else
      l.push_back(*si);
  return symbol(l);
}


// Read/write code

inline static bool dont_escape(char c) { 
  return isalnum(c) || c == '_' || c == '.' || c == '-' || c == '+'; 
}

inline static char escaped_char(char c) {
  switch (c) {
  case 'a': return('\a');
  case 'b': return('\b');
  case 'f': return('\f');
  case 'n': return('\n');
  case 'r': return('\r');
  case 't': return('\t');
  case 'v': return('\v');
  default: return c;
  }
  return c;
}

std::istream& operator>> (std::istream& is, symbol& s)
{
  std::string str;
  char c;
  if (!(is >> c)) return is;           // If read fails, return error
  if (dont_escape(c) || c == ESCAPE) { // Recognize a normal symbol
    do {
      if (c == ESCAPE) {
	if (!is.get(c)) return is;     //  Read next character; return if read fails.
	str.push_back(escaped_char(c));//  Push escaped char onto string.
      }
      else
	str.push_back(c);
    }
    while (is.get(c) && (dont_escape(c) || c == ESCAPE));
    if (!is.fail())                    //  Did we read one too many chars?
      is.putback(c);                   //   Yes.  Put it back.
    else
      is.clear(std::ios::failbit);     //   No.  We're probably at end-of-file.
    s = symbol(str);                   //  Load string into symbol
  }
  else if (c == OPENQUOTE) {           // Recognize a quoted string
    if (!is.get(c)) return is;         //  Read next character; return if read fails
    while (c != CLOSEQUOTE) {
      if (c == ESCAPE) {               //  Is this character the escape character?
	if (!is.get(c)) return is;     //   Yes.  Get quoted character.
	str.push_back(escaped_char(c));//   Push character onto string.
      }
      else
	str.push_back(c);              //   Push back ordinary character.
      if (!is.get(c)) return is;       //  Read next character.
    }
    s = symbol(str);                   //  Load string into symbol
  }
  else if (c == UNDEFINED[0]) {
    for (const char* cp = &UNDEFINED[1]; *cp; ++cp)
      if (!is.get(c) || c != *cp) {
	is.clear(std::ios::failbit);   //  We didn't get the whole UNDEFINED symbol
	return is;
      }
    s = symbol::undefined();           //  Set s to undefined
  }
  else {                               // c doesn't begin a symbol
    is.putback(c);                     // put it back onto the stream
    is.clear(std::ios::failbit);       // set the fail bit
  }
  return is;
}


std::ostream& operator<< (std::ostream& os, const symbol s)
{
  if (s.is_undefined())
    os << UNDEFINED;
  else {
    const std::string& str = s.string_reference();
    if (str.empty())
      os << OPENQUOTE << CLOSEQUOTE;
    else
      for (std::string::const_iterator si = str.begin(); si != str.end(); ++si) {
	if (!dont_escape(*si))
	  os.put(ESCAPE);
	os.put(*si);
      }
  }
  return os;
}
