// transitive-verbs.h
//
// Mark Johnson, 11th December 2001
//
// This code counts how often each verb in the training corpus is
// followed by an NP, in an effort to work out whether the verb is
// usually transitive or not.

#ifndef TRANSITIVE_VERBS_H
#define TRANSITIVE_VERBS_H

#include "sym.h"
#include "symset.h"
#include "transitive-verbs.h"
#include "tree.h"
#include "utility.h"

#include <cassert>
#include <cctype>
#include <ext/hash_map>
#include <ext/hash_set>
#include <iostream>
#include <string>
#include <vector>

typedef std::vector<tree*> Ts;

static const symset verb_pos("AUX AUXG VB VBD VBG VBN VBP VBZ");
static const symbol NP("NP");

struct TransitiveVerbs {

public:
  typedef ext::hash_map<symbol,size_t> S_I;
  typedef ext::hash_set<symbol> sS;
  
  sS transitive_verbs;

private:

  struct CountVerbs {
    S_I& verb_count;
    S_I& transverb_count;

    CountVerbs(S_I& verb_count, S_I& transverb_count) :
      verb_count(verb_count), transverb_count(transverb_count) { }

    void operator() (const tree* t) {
      if (verb_pos.contains(t->label.cat)) {
	assert(t->is_preterminal());
	symbol verb = t->child->label.cat.to_lower();
	++verb_count[verb];
	if (t->next && t->next->label.cat == NP)
	  ++transverb_count[verb];
      }
    }
  };

public:
  void load_transitive_verbs(const Ts& tps) {
    S_I verb_count;
    S_I transverb_count;

    CountVerbs cv(verb_count, transverb_count);
    cforeach (Ts, tpi, tps)
      (*tpi)->preorder(cv);

    cforeach (S_I, i, transverb_count) {
      size_t trans_count = i->second;
      size_t count = afind(verb_count, i->first);
      if (2*trans_count >= count) 
	transitive_verbs.insert(i->first);
    }
  }
  
  bool transitive_verb(symbol pos, symbol verb) {
    if (verb_pos.contains(pos)) 
      return (transitive_verbs.count(verb.to_lower()) == 1);
    else 
      return false;
  }

  void label_transitive_verbs(tree* t) {
    assert(t);
    if (t->is_preterminal()) { 
      symbol pos = t->label.cat;
      if (transitive_verb(pos, t->child->label.cat)) 
	t->label.cat = symbol(pos.string_reference() + "_t");
    }
    else
      label_transitive_verbs(t->child);
    if (t->next)
      label_transitive_verbs(t->next);
  }

  void label_transitive_verbs(Ts& ts) {
    foreach (Ts, i, ts)
      label_transitive_verbs(*i);
  }

  void remove_transitive_verb_labels(tree* t) {
    assert(t);
    if (t->is_preterminal()) { 
      const std::string& pos = t->label.cat;
      std::string::size_type i = pos.find('_');
      if (i != std::string::npos) {
	std::string pos_root = pos.substr(0, i);
	t->label.cat = symbol(pos_root);
      }
    }
    else
      remove_transitive_verb_labels(t->child);
    if (t->next)
      remove_transitive_verb_labels(t->next);
  }
};  // TransitiveVerbs{ }

inline std::ostream& operator<< (std::ostream& os, const TransitiveVerbs& tv)
{
  return os << tv.transitive_verbs << std::endl;
}

inline std::istream& operator>> (std::istream& is, TransitiveVerbs& tv)
{
  tv.transitive_verbs.clear();
  return is >> tv.transitive_verbs;
}

#endif // TRANSITIVE_VERBS_H
