// tree.h
//
// (c) Mark Johnson, 20th August 2001, 
// modified (c) 15th September 2001
// modified (c) 19th January 2002
//
// The tree structure can represent arbitrary trees.
// Each tree node contains:
// - the node's label
// - a pointer to its first child (in subtrees)
// - a pointer to the right sibling of this node
//
// IMPORTANT NOTE: there is no automatic memory management
// for tree nodes.  readtree_root returns a tree pointer.
// You should manually delete the tree pointer when you are
// done with it.  Use delete_tree to do this.
//
// map_filenamefile_trees always deletes each tree after it
// has been read, so in this case you don't have to worry about
// deleting the trees.
//
// read_filenamefile_trees loads a vector with all of the trees
// in the corpus.  If you decide to delete this vector you should
// manually delete the tree points also.  Use delete_tree on
// each pointer in order to do this.



#ifndef TREE_H
#define TREE_H

#include "sym.h"
#include "utility.h"

#include <algorithm>
#include <cassert>
#include <cstdio>
#include <deque>
#include <fstream>
#include <iostream>
#include <memory>
#include <set>
#include <sstream>
#include <string>
#include <vector>

extern int readtree_lineno;              // defined in read-tree.l
extern const char* readtree_filename;

///////////////////////////////////////////////////////////////////////////
//                                                                       //
//                                tree_label                             //
//                                                                       //
///////////////////////////////////////////////////////////////////////////

struct tree_label {
  typedef symbol cat_type;
  typedef size_t index_type;

  cat_type   cat;   // category
  index_type index; // index (0 = no index)

  tree_label(const symbol cat_ = symbol::undefined(), size_t index_ = 0)
    : cat(cat_), index(index_) { };
};

namespace std {   // these are needed for map and hash
  template <> struct equal_to<tree_label> 
    : public binary_function<tree_label,tree_label,bool> {
    bool operator() (const tree_label& l1, const tree_label& l2) const {
      return equal_to<symbol>()(l1.cat, l2.cat) && l1.index == l2.index;
    }
  };

  template <> struct less<tree_label>
    : public binary_function<tree_label,tree_label,bool> {
    bool operator() (const tree_label& l1, const tree_label& l2) const {
      return less<symbol>()(l1.cat, l2.cat) || (!less<symbol>()(l2.cat,l1.cat) && l1.index < l2.index);
    }
  };

  template <> struct greater<tree_label>
    : public binary_function<tree_label,tree_label,bool> {
    bool operator() (const tree_label& l1, const tree_label& l2) const {
      return greater<symbol>()(l1.cat, l2.cat) || (!greater<symbol>()(l2.cat,l1.cat) && l1.index > l2.index);
    }
  };

};  // namespace std

namespace EXT_NAMESPACE {
  template <> struct hash<tree_label> {
    size_t operator()(const tree_label& l) const {
      return hash<symbol>()(l.cat)+l.index;
    }
  };
}; // namespace EXT_NAMESPACE

std::ostream& operator<< (std::ostream& os, const tree_label& label);


///////////////////////////////////////////////////////////////////////////
//                                                                       //
//                                   tree                                //
//                                                                       //
///////////////////////////////////////////////////////////////////////////

template <class label_type_=tree_label> class tree_node {
public:
  typedef label_type_ label_type;

  label_type label;

  tree_node* child;   // subtrees
  tree_node* next;    // sibling

  tree_node(tree_node* child_ = NULL, tree_node* next_ = NULL) 
    : child(child_), next(next_) { }

  tree_node(const label_type& label_, tree_node* child_ = NULL, tree_node* next_ = NULL) 
    : label(label_), child(child_), next(next_) { }

  tree_node(typename tree_label::cat_type cat_, typename tree_label::index_type index_ = 0, 
	    tree_node* child_ = NULL, tree_node* next_ = NULL) 
    : label(tree_label::tree_label(cat_, index_)), child(child_), next(next_) { }

  ~tree_node() { delete child; delete next; }


  bool is_terminal() const {
    return child == NULL;
  }

  bool is_preterminal() const {
    return child && child->is_terminal();
  }

  bool is_nonterminal() const {
    return child && child->child;
  }

  void delete_this() { child = next = NULL; delete this; }

  // copy_tree() returns a pointer to a ``deep copy'' of this node
  //  and all of the nodes it points to
  //
  tree_node* copy_tree() const {
    return new tree_node(label, 
			 child != NULL ? child->copy_tree() : NULL,
			 next != NULL ? next->copy_tree() : NULL);
  }

  // preorder_find() returns the first node for which pred is true
  //
  template <typename Pred> const tree_node* preorder_find(Pred pred) const
  {
    if (pred(this))
      return this;
    if (child) 
      if (const tree_node* rc = child->preorder_find(pred))
	return rc;
    if (next)
      if (const tree_node* rn = next->preorder_find(pred))
	return rn;
    return NULL;
  }

  // count() counts the number of nodes on which pred is true
  //
  template <typename Pred> size_t count(Pred pred, size_t cnt = 0) const
  {
    if (pred(this))
      ++cnt;
    if (child)
      cnt = child->count(pred, cnt);
    return next ? next->count(pred, cnt) : cnt;
  }

private:
  // helper function for size()
  //
  size_t size_(size_t n) const
  {
    ++n;
    if (child)
      n = child->size_(n);
    return next ? next->size_(n) : n;
  }

public:
  // size() returns the number of nodes in tree
  //
  size_t size() const
  {
    return size_(0);
  }


  // max_depth() returns the length of the longest path of nodes to a terminal
  //  all siblings are at same depth, children are at depth 1, etc.
  //
  size_t max_depth() const
  {
    return std::max(child ? 1+child->max_depth() : 0,
		    next ? next->max_depth() : 0);
  }

  // is_empty() is true iff all terminals have POS -NONE-
  //
  bool is_empty() const
  {
    if (label.cat == symbol::none)
      return true;
    if (!child)
      return false;
    for (const tree_node* c = child; c; c = c->next)
      if (!c->is_empty())
	return false;
    return true;
  }

  // is_adjunction_site() is true iff all but one child is empty(), and that child
  //  has the same label as this node
  //
  bool adjunction_site() const
  {
    if (!child)
      return false;

    int non_empty_same(0), empty(0);

    for (const tree_node* c = child; c; c = c->next) {
      if (c->is_empty()) 
	++empty;
      else {
	if (c->label.cat != label.cat)
	  return false;
	else {
	  if (non_empty_same == 0) 
	    ++non_empty_same;
	  else
	    return false;
	}}}

    if (empty == 0) {
      assert(!(child->child || child->next)); // Unary branching node with same category
      return false;
    }
    
    return non_empty_same == 1;
  }


  void delete_unary_same_label_chains() { // ignores terminals
    if (child) {
      while (child->child && !child->next && child->label.cat == label.cat) {
 	tree_node* grandchild = child->child;
 	child->delete_this();
 	child = grandchild;
      }
      child->delete_unary_same_label_chains();
    }
    if (next)
      next->delete_unary_same_label_chains();
  }


  // copy_without_empties() returns a ptr to a copy of this subtree with all
  //  empty nodes and indicies deleted
  //
  tree_node* copy_without_empties(tree_node* endp = NULL) const {
    if (is_empty()) 
      return next ? next->copy_without_empties(endp) : endp;
    else if (adjunction_site()) {
      assert(child);
      return child->copy_without_empties(next ? next->copy_without_empties(endp) : endp);
    }
    else {
      tree_node* c = new tree_node(label, 
				   child ? child->copy_without_empties() : NULL, 
				   next ? next->copy_without_empties(endp) : endp);
      c->label.index = 0;
      return c;
    }
  }


  // copy_left_binarize() returns a pointer to a left binarized copy of
  //  this tree
  //
  tree_node* copy_left_binarize() const {
    tree_node* next_copy = NULL;
    if (next) {
      next_copy = next->copy_left_binarize();
      if (next_copy->next)
	next_copy = new tree_node(label_type(), next_copy, NULL);
    }
    return new tree_node(label,
			 child ? child->copy_left_binarize() : NULL,
			 next_copy);
  }

  // copy_parent_annotate() returns a pointer to a parent-annotated 
  //  copy of this tree
  //
  tree_node* copy_parent_annotate(symbol parent_cat=symbol::undefined()) const {
    tree_node* t = new tree_node(label);
    assert(child);
    if (is_preterminal() || is_empty()) 
      t->child = child->copy_tree();
    else {
      t->child = child->copy_parent_annotate(label.cat);
      if (parent_cat.is_defined())
	t->label.cat = symbol(label.cat.string_reference() + "_" + parent_cat.string_reference());
    }
    if (next)
      t->next = next->copy_parent_annotate(parent_cat);
    return t;
  }

  template <typename Fn> void preorder(Fn& fn) {
    fn(this);
    if (child) 
      child->preorder(fn);
    if (next)
      next->preorder(fn);
  }

  template <typename Fn> void preorder(Fn& fn) const {
    fn(this);
    if (child) 
      child->preorder(fn);
    if (next)
      next->preorder(fn);
  }

  template <typename Fn> void postorder(Fn& fn) {
    if (child) 
      child->postorder(fn);
    fn(this);
    if (next)
      next->postorder(fn);
  }

  template <typename Fn> void postorder(Fn& fn) const {
    if (child) 
      child->postorder(fn);
    fn(this);
    if (next)
      next->postorder(fn);
  }

  template <typename Fn> bool find(Fn fn) const {
    return fn(this) || (child && child->find(fn)) || (next && next->find(fn));
  }


private:
  template <typename Fn> void preorder_ancestors(Fn& fn, std::vector<const tree_node*>& ancestors)  
  {
    ancestors.push_back(this);
    fn(ancestors);
    if (child) {
      child->preorder_ancestors(fn, ancestors);
    }
    ancestors.pop_back();
    if (next)
      next->preorder_ancestors(fn, ancestors);
  }

public:

  // preorder_ancestors() calls fn(as) where as.back() is the 
  //  node being visited, as[0] is the root node and as[i-1] is
  //  the parent of as[i].
  //
  template <typename Fn> void preorder_ancestors(Fn& fn) 
  {
    std::vector<const tree_node*> ancestors;
    preorder_ancestors(fn, ancestors);
    assert(ancestors.empty());
  }

  // terminals() appends the non-empty terminals to terms
  //
  template <typename Ss> void terminals(Ss& terms) const
  {
    if (child == NULL) 
      terms.push_back(label.cat);
    else { // child != NULL
      if (label.cat != symbol::none)
	child->terminals(terms);
    }
    if (next != NULL)
      next->terminals(terms);
  }

  // fillers_gaps_indices() inserts all filler indices into fillers
  //  and all gap indices into gaps
  //
  template <typename Set> void fillers_gaps_indices(Set& fillers, Set& gaps) const
  {
    if (label.cat == symbol::none) {
      assert(child);
      if (child->label.index)
	gaps.insert(child->label.index);
    }
    else {
      if (label.index)
	fillers.insert(label.index);
      if (child)
	child->fillers_gaps_indices(fillers, gaps);
    }
    if (next)
      next->fillers_gaps_indices(fillers, gaps);
  }

  // fillers_equal_gaps() checks that the indices on the fillers and 
  //  the gaps in the tree match
  //
  bool fillers_equal_gaps() const {
    std::set<size_t> fillers, gaps;
    fillers_gaps_indices(fillers, gaps);
    return fillers == gaps;
  }

  // fillers_include_gaps() checks that every indexed gap in the tree
  //  has an indexed filler
  //
  bool fillers_include_gaps() const {
    std::set<size_t> fillers, gaps;
    fillers_gaps_indices(fillers, gaps);
    return includes(fillers, gaps);
  }
};  // class tree_node


namespace std {
  template <typename label_type> struct equal_to<tree_node<label_type> >
    : public binary_function<tree_node<label_type>, tree_node<label_type>, bool> {
    bool operator() (const tree_node<label_type>& t1, const tree_node<label_type>& t2) const {
      return equal_to<label_type>()(t1.label, t2.label) 
	     && equal_to<tree_node<label_type>* >()(t1.next, t2.next)
	     && equal_to<tree_node<label_type>* >()(t1.child, t2.child);
    }
  };

  template <typename label_type> struct less<tree_node<label_type> >
    : public binary_function<tree_node<label_type>, tree_node<label_type>, bool> {
    bool operator() (const tree_node<label_type>& t1, const tree_node<label_type>& t2) const {
      if (less<label_type>()(t1.label, t2.label)) return true;
      if (less<label_type>()(t2.label, t1.label)) return false;
      if (less<tree_node<label_type>*>()(t1.child, t2.child)) return true;
      if (less<tree_node<label_type>*>()(t2.child, t1.child)) return false;
      return less<tree_node<label_type>*>()(t1.next, t2.next);
    }
  };

  template <typename label_type> struct greater<tree_node<label_type> >
    : public binary_function<tree_node<label_type>, tree_node<label_type>, bool> {
    bool operator() (const tree_node<label_type>& t1, const tree_node<label_type>& t2) const {
      if (greater<label_type>()(t1.label, t2.label)) return true;
      if (greater<label_type>()(t2.label, t1.label)) return false;
      if (greater<tree_node<label_type>*>()(t1.child, t2.child)) return true;
      if (greater<tree_node<label_type>*>()(t2.child, t1.child)) return false;
      return greater<tree_node<label_type>*>()(t1.next, t2.next);
    }
  };

  template <typename label_type> struct equal_to<tree_node<label_type>*> 
    : public binary_function<tree_node<label_type>*,tree_node<label_type>*,bool> {
    bool operator() (const tree_node<label_type>* t1, const tree_node<label_type>* t2) const {
      if (!t1 && !t2) return true;
      if (!t1 || !t2) return false;
      if (t1 == t2) return true;
      return equal_to<tree_node<label_type> >()(*t1, *t2);
    }
  };

  template <typename label_type> struct less<tree_node<label_type>*>
    : public binary_function<tree_node<label_type>*,tree_node<label_type>*,bool> {
    bool operator() (const tree_node<label_type>* t1, const tree_node<label_type>* t2) const {
      if (!t1 && !t2) return false;
      if (!t1 && t2) return true;
      if (!t2 && t1) return false;
      if (t1 == t2) return false;
      return less<tree_node<label_type> >()(*t1, *t2);
    }
  };
}  // namespace std

namespace EXT_NAMESPACE {
  template <typename label_type> struct hash<tree_node<label_type> > {
    typedef tree_node<label_type> tree;
    size_t operator()(const tree& t) const {
      return 2*hash<label_type>()(t.label) + 3*hash<tree*>()(t.next) + 5*hash<tree*>()(t.child);
    }
  };

  template <typename label_type> struct hash<tree_node<label_type>*> {
    typedef tree_node<label_type> tree;
    size_t operator()(const tree* t) const {
      if (t) 
	return hash<tree>()(*t);
      else
	return 0;
    }
  };
}; // namespace EXT_NAMESPACE

template <typename label_type> 
inline void delete_tree(tree_node<label_type>* t) { 
  delete t; 
}


typedef tree_node<> tree;

tree* readtree_root(FILE* fp);  // hallucinate a ROOT node label, needed for Penn TB trees
tree* readtree(FILE* fp);       // don't hallucinate a ROOT node label, needed for BLLIP trees

void read_filenamefile_trees(const char* filenamefile, std::vector<tree*>& trees, 
			     tree* (*reader)(FILE*) = &readtree_root);


template<typename Proc>
void map_filenamefile_trees(const char* filenamefile, Proc& proc)
{
  std::ifstream is(filenamefile);
  if (!is) {
    std::cerr << "Error in tree.h read_tree_file(): can't open filenamefile " 
	      << filenamefile << std::endl;
    abort();
  }

  std::string filename;
  while (is >> filename) {
    FILE* fp = fopen(filename.c_str(), "r");
    if (fp == NULL) {
      std::cerr << "Error in tree.h read_tree_file(): can't open filename " 
		<< filename << std::endl;
      abort();
    }

    readtree_filename = filename.c_str();
    readtree_lineno = 1;

    std::auto_ptr<tree> tp;
    while ((tp = readtree_root(fp)).get()) 
      proc(tp.get());

    fclose(fp);
  }
}

template <typename label_type>
static void write_tree(std::ostream& s, const tree_node<label_type>* t)
{
  assert(t);

  if (t->child) {
    s << '(' << t->label;
    
    for (tree_node<label_type> *p = t->child; p; p = p->next) {
      s << ' ';
      write_tree(s, p);
    }

    s << ')';
  }
  else 
    s << t->label;
}

template <typename label_type>
std::ostream& operator<< (std::ostream& os, const tree_node<label_type>* t) {
  if (t)
    write_tree(os, t);
  else
    os << "_NULL_";
  return os;
}

template <typename label_type>
static void write_prolog_tree_(std::ostream& fp, const tree_node<label_type>* t)
{
  fp << t->label;
  if (t->child) {
    fp << "/[";
    
    for (tree* p = t->next; p; p = p->next) {
      write_prolog_tree_(fp, p);
      
      if (p->next)
	fp << ',';
    }

    fp << ']';
  }
}

template <typename label_type>
void write_prolog_tree(std::ostream& fp, const tree_node<label_type>* t)
{
  write_prolog_tree_(fp, t);
  fp << '.' << std::endl;
}

template <typename T>
size_t print_length(const T& t)
{
  std::ostringstream os;
  os << t;
  return os.str().size();
}


template <typename label_type>
void display_tree(std::ostream& fp, const tree_node<label_type>* t, int indent=0)
{
  if (t->child) {
    fp << '(' << t->label << ' ';
    indent += print_length(t->label) + 2;  // for '(' and ' '
    display_tree(fp, t->child, indent);

    for(tree_node<label_type>* p = t->child->next; p; p = p->next) {
      fp << '\n';
      for (int i=0; i<indent; i++)
	fp << ' ';
      display_tree(fp, p, indent);
    }
    fp << ')';
  }
  else 
    fp << t->label;
}


// collapses non-branching unary chains -- only collapses nonterms, not preterms
//
// template <typename tree_label>
// tree_node<tree_label>* collapse_identical_unary_nonterms(const tree_node<tree_label>* t)
// {
//   typedef tree_node<tree_label> tree;

//   if (!t) 
//     return NULL;
//   else {
//     size_t index = t->label.index;
//     tree* p = t->child;
//     while (p && p->child && !p->next && p->label.cat == t->label.cat) {
//       assert(!(index && p->label.index));  // both parent and child don't have an index
//       if (!index && p->label.index)
// 	index = p->label.index;
//       p = p->child;
//     }
//     return new tree(tree_label(t->label.cat, index), 
// 		    collapse_identical_unary_nonterms(p), 
// 		    collapse_identical_unary_nonterms(t->next));
//   }}

// clears and then loads terms with the terminal categories of t
void tree_terms(const tree* t, std::vector<symbol>& terms);





// tree_preterms
//
template <typename label_type>
struct tree_preterms_  // function object helper class
{
  typedef tree_node<label_type> tree;

  std::vector<symbol>* preterms;

  tree_preterms_(std::vector<symbol>& preterms_) : preterms(&preterms_) { }

  void operator() (const tree* t, const std::deque<const tree*>& parents) {
    assert(t->child != NULL || !parents.empty());
    if (t->child == NULL && parents.front()->label.cat != symbol::none)
      preterms->push_back(parents.front()->label.cat);
  }
};

// tree_preterms 
//  clears and then loads preterms with the preterminal categories of t
//
template <typename label_type>
void tree_preterms(const tree_node<label_type>* t, std::vector<symbol>& preterms)
{     
  preterms.clear();
  tree_preterms_<label_type> tp(preterms);
  t->preorder_parents(tp);
}

#endif
